/**
 * $Id: SalsaCompiler.java,v 1.8 2002/08/09 20:06:42 stepha Exp $
 *
 * This is the Salsa programming language compiler.  Compilation goes
 * through two major steps:
 *
 * (1) Parsing phase:  Here we call the Salsa Parser.
 *                     As generated by JavaCC with a few changes to keep
 *                     the input for code generation.
 *
 * (2) Code Generation phase:  Here we use the abstract syntax tree
 *                             produced by the first phase to build 
 *                             Java files corresponding to the original
 *                             Salsa files being compiled.
 * Modified by Robin D. Toll	2001/12
 *		-Re-coded parameter handling.
 *
 * @author Carlos Varela   cvarela@uiuc.edu   1999/09/14
 * @version 0.3.3 2002/04/02
 */

import java.io.*;

public class SalsaCompiler {
    public static final String VERSION = "0.4.0";
    private String[] files = null;

	public static void main(String args[]) {
		new SalsaCompiler(args);
	}

	SalsaCompiler(String[] args) {
		SalsaParser parser = null;
		SalsaCodeGenerator codeGenerator;
		String outputFile = null;

		if (args.length == 0) {
			System.out.println("Salsa Compiler Version "+
					VERSION+
					":  Reading from standard input . . .");
			parser = new SalsaParser(System.in);
		} else {
			for (int x=0; x<args.length; x++) {
				if (args[x].equals("-?"))
					help();
				else
					addFile(args[x]);
			}
			if ((files == null) || (files.length < 1)) {
				System.out.println("Salsa Compiler Version "+
						VERSION+
						":  Reading from standard input . . .");
				parser = new SalsaParser(System.in);
				//help();
			}
//			System.out.println(files.length+" files to process.");
			for (int x=0; x<files.length; x++) {
				System.out.println();
				System.out.println("Salsa Compiler Version "+
						VERSION+
						":  Reading from file " + files[x] + " . . .");
				try {
					if (x==0) {
						parser = new SalsaParser(new java.io.FileInputStream(files[x]));
					} else {
						parser.ReInit(new java.io.FileInputStream(files[x]));
					}
					if (!(files[x].endsWith(".salsa"))) {
						System.out.println("Salsa Compiler Version "
								+VERSION
								+":  File must end with \".salsa\"");
					} else { // Okay, it's passed all our tests; lets actually compile it.
						outputFile = (files[x].substring(0, files[x].length()-6).concat(".java"));
					}
					try {
						SalsaNode n = parser.CompilationUnit();
						System.out.println("Salsa Compiler Version "+
								VERSION+
								":  Salsa program parsed successfully.");
						if (outputFile != null)
						// cvarela: if it exists, it should rename it...
							codeGenerator = new SalsaCodeGenerator(n, new FileOutputStream(outputFile));
						else
							codeGenerator = new SalsaCodeGenerator(n, System.out);
						codeGenerator.generateCode();
						System.out.println("Salsa Compiler Version "+
								VERSION+
								":  Salsa program compiled successfully.");
					} catch (IOException e) {
						System.out.println(e.getMessage());
						System.out.println("Salsa Compiler Version "+
								VERSION+
								":  Encountered errors writing to file:"+outputFile);
					} catch (ParseException e) {
						System.out.println(e.getMessage());
						System.out.println("Salsa Compiler Version "+
								VERSION+
								":  Encountered errors during parse.");
					} catch (CodeGenerationException e) {
						System.out.println(e.getMessage());
						System.out.println("Salsa Compiler Version "+
								VERSION+
								":  Encountered errors during code generation.");
					}
				} catch (java.io.FileNotFoundException e) {
					System.out.println("Salsa Compiler Version "+
							VERSION+
							":  File " + files[x] + " not found.");
				}

			}
		}
/*	removed 0.3.1	--RDT
		if (args.length == 0) {
			System.out.println("Salsa Compiler Version "+
					VERSION+
					":  Reading from standard input . . .");
			parser = new SalsaParser(System.in);
		} else if (args.length == 1) {
			System.out.println("Salsa Compiler Version "+
					VERSION+
					":  Reading from file " + args[0] + " . . .");
			try {
				parser = new SalsaParser(new java.io.FileInputStream(args[0]));
				if (args[0].endsWith(".salsa")){
					outputFile = (args[0].substring(0,args[0].length()-6).concat(".java"));
				} else {
					System.out.println("Salsa Compiler Version "+
							VERSION+
							":  File must end with \".salsa\"");
					return;
				}
			} catch (java.io.FileNotFoundException e) {
				System.out.println("Salsa Compiler Version "+
						VERSION+
						":  File " + args[0] + " not found.");
				return;
			}
		} else {
			System.out.println("Salsa Compiler Version "+
					VERSION+
					":  Usage is one of:");
			System.out.println("         java SalsaCompiler < inputfile");
			System.out.println("OR");
			System.out.println("         java SalsaCompiler inputfile");
			return;
		}
		try {
			SalsaNode n = parser.CompilationUnit();
			System.out.println("Salsa Compiler Version "+
					VERSION+
					":  Salsa program parsed successfully.");
			if (outputFile != null)
			// cvarela: if it exists, it should rename it...
				codeGenerator = new SalsaCodeGenerator(n, new FileOutputStream(outputFile));
			else codeGenerator = new SalsaCodeGenerator(n, System.out);
				codeGenerator.generateCode();
			System.out.println("Salsa Compiler Version "+
					VERSION+
					":  Salsa program compiled successfully.");
		} catch (IOException e) {
			System.out.println(e.getMessage());
			System.out.println("Salsa Compiler Version "+
					VERSION+
					":  Encountered errors writing to file:"+outputFile);
		} catch (ParseException e) {
			System.out.println(e.getMessage());
			System.out.println("Salsa Compiler Version "+
					VERSION+
					":  Encountered errors during parse.");
		} catch (CodeGenerationException e) {
			System.out.println(e.getMessage());
			System.out.println("Salsa Compiler Version "+
					VERSION+
					":  Encountered errors during code generation.");
		}
*/
	}

	private static void help() {
		System.out.println("Salsa Compiler Version "+
				VERSION+
				":  Usage is one of:");
		System.out.println("         java SalsaCompiler < inputfile");
		System.out.println("OR");
		System.out.println("         java SalsaCompiler inputfile");
		System.exit(1);
	}

	private void addFile(String fn) {
		if (files == null) {
			files = new String[1];
			files[0] = fn;
		} else {
			String[] tmp = new String[files.length+1];
			for (int x=0; x<files.length; x++)
				tmp[x] = files[x];
			tmp[files.length] = fn;
			files = tmp;
		}
	}
}
