/**
 * $Id: SalsaNode.java,v 1.19 2002/08/21 09:21:28 deselt Exp $
 *
 * A SalsaNode instance contains a node of the abstract syntax tree of
 * a Salsa program already parsed.  It also contains the necessary behavior
 * to convert the node into Java code.
 *
 * Author: Carlos Varela   cvarela@uiuc.edu   1999/09/29.
 */
public class SalsaNode extends SimpleNode {
   // These tokens allow us to reconstruct the input from Salsa programs
   Token first_token, last_token;

   // These hashtables tell us what code may our children need to include
   // for particular NonTerminal nodes.
   java.util.Hashtable preTokens  = new java.util.Hashtable(3);
   java.util.Hashtable postTokens = new java.util.Hashtable(3);

   // This flag tells us whether or not we're generating code for
   // message sending.
   boolean messageSend;

   // This flag tell us whether or not to generate code for this node
   boolean generateCode = true;

   public SalsaNode(int id) {
      super(id);
   }

   public SalsaNode(SalsaParser p, int id) {
      super(p, id);
   }

   public Token getFirstToken() {
      return first_token;
   }

   public Token getLastToken() {
      return last_token;
   }

   public int getId() {
      return id;
   }

   public String image() {
      StringBuffer tokenString = new StringBuffer();
      Token token = first_token;
      do {
         tokenString.append(token.image);
         if (token == last_token) break;
            token = token.next;
      } while (token!=null);
      return new String(tokenString);
   }

   public String toString() {
      return (SalsaParserTreeConstants.jjtNodeName[id]+" "+image());
   }

   public String toString(String prefix) {
      return prefix + toString();
   }

   public void dump(String prefix) {
      java.io.PrintWriter stdOut = new java.io.PrintWriter(System.out);
      dump(stdOut, prefix);
      stdOut.flush();
   }

   public void dump(java.io.PrintWriter writer, String prefix) {
      writer.println(toString(prefix));
      if (children != null) {
         for (int i = 0; i < children.length; ++i) {
            SalsaNode n = (SalsaNode)children[i];
            if (n != null) {
               n.dump(writer, prefix + " ");
            }
         }
      }
   }

   public String preJavaCode() throws CodeGenerationException {
      StringBuffer tokenString = new StringBuffer(preToken(id));
      switch (id) {
         case SalsaParserTreeConstants.JJTCOMPILATIONUNIT:
            return ("/* Generated by Salsa Compiler v"
                  +SalsaCompiler.VERSION+" */\n");
         case SalsaParserTreeConstants.JJTMODULEDECLARATION:
            return ("\npackage ");
         case SalsaParserTreeConstants.JJTIMPORTDECLARATION:
            return ("\nimport ");
         case SalsaParserTreeConstants.JJTUNMODIFIEDBEHAVIORDECLARATION:
            String behaviorDeclaration = "\npublic class "+first_token.next.image;
            SalsaCodeGenerator.setBehaviorName(first_token.next.image);
            if (jjtGetNumChildren() > 1) {
               inToken(" extends ", SalsaParserTreeConstants.JJTNAME);
               inToken(" implements ", SalsaParserTreeConstants.JJTNAMELIST);
               return behaviorDeclaration;
            } else
               return behaviorDeclaration+" extends UniversalActor";
         case SalsaParserTreeConstants.JJTBEHAVIORBODY:
            behaviorDeclaration = "{\n"
               +"\n\tstatic StandardOutput standardOutput"
               +" = StandardOutput.getInstance();\n"
               +"\n\tstatic StandardError standardError"
               +" = StandardError.getInstance();\n"

               +"\n\tpublic " + SalsaCodeGenerator.behaviorName + " selfReference;\n";

               // Determine if this inner class should extend somethind else.
               if (((SalsaNode)parent).jjtGetNumChildren() > 1) {
                   String extendsName = ((SalsaNode)parent.jjtGetChild( 0 )).getFirstToken().image;

                   behaviorDeclaration +=
                  "\n\tpublic class State extends " + extendsName + ".State {\n\n";
               } else {
                   behaviorDeclaration +=
                  "\n\tpublic class State extends UniversalActorState {\n\n";
               }
  
		behaviorDeclaration += "\n\t\tprotected " + SalsaCodeGenerator.behaviorName + " self;\n\n";

	       return behaviorDeclaration;

         case SalsaParserTreeConstants.JJTBEHAVIORBODYDECLARATION:
            return "";
         case SalsaParserTreeConstants.JJTBLOCK:
            return ("{\n");
         case SalsaParserTreeConstants.JJTUNMODIFIEDINTERFACEDECLARATION:
            return "\npublic interface "+first_token.next.image+" ";
         case SalsaParserTreeConstants.JJTFIELDDECLARATION:
            return "\t\t";
         case SalsaParserTreeConstants.JJTVARIABLEDECLARATOR:
            inToken("=", SalsaParserTreeConstants.JJTVARIABLEINITIALIZER);
            return("");
         case SalsaParserTreeConstants.JJTARRAYINITIALIZER:
            return ("{");
         case SalsaParserTreeConstants.JJTMETHODDECLARATION:
            // cvarela:  need to consider private/protected/etc
            return ("\n\t\tpublic ");
         case SalsaParserTreeConstants.JJTVARIABLEDECLARATORID:
         case SalsaParserTreeConstants.JJTPRIMITIVETYPE:
         case SalsaParserTreeConstants.JJTNAME:
         case SalsaParserTreeConstants.JJTASSIGNMENTOPERATOR:
         case SalsaParserTreeConstants.JJTSTRINGLITERAL:
         case SalsaParserTreeConstants.JJTNULLLITERAL:
         case SalsaParserTreeConstants.JJTINTEGERLITERAL:
         case SalsaParserTreeConstants.JJTFLOATINGPOINTLITERAL:
         case SalsaParserTreeConstants.JJTCHARACTERLITERAL:
         case SalsaParserTreeConstants.JJTBOOLEANLITERAL:
            return image();
         case SalsaParserTreeConstants.JJTMETHODDECLARATOR:
            if (first_token.image.equals("act"))
               SalsaCodeGenerator.generateMain = true;
            return (first_token.image);
         case SalsaParserTreeConstants.JJTFORMALPARAMETERS:
            if (jjtGetNumChildren() > 1) {
               inPostToken(", ", SalsaParserTreeConstants.JJTFORMALPARAMETER);
            }
            return ("(");
         case SalsaParserTreeConstants.JJTCONSTRUCTORDECLARATION:

            inToken("throws ", SalsaParserTreeConstants.JJTNAMELIST);
            inToken("{\n", SalsaParserTreeConstants.JJTEXPLICITCONSTRUCTORINVOCATION);
            inToken("{\n", SalsaParserTreeConstants.JJTBLOCKSTATEMENT);
            // return ("\t\t"+first_token.image);

            return "\t\tState ";
         case SalsaParserTreeConstants.JJTEXPLICITCONSTRUCTORINVOCATION:
            switch (first_token.kind) {
               case SalsaParserConstants.THIS:
                  return ("this");
               case SalsaParserConstants.SUPER:
                  return ("super");
               default:
                  // PrimaryExpression
                  inToken(".super", SalsaParserTreeConstants.JJTARGUMENTS);
                  return ("");
            }
         case SalsaParserTreeConstants.JJTRESULTTYPE:
            if (first_token.kind == SalsaParserConstants.VOID)
               return ("void");
            else
               return ("");
         case SalsaParserTreeConstants.JJTNAMELIST:
            // Modified by Travis Desell,  May 6, 2002

            if (jjtGetNumChildren() > 1)
               for (int i = 1; i < jjtGetNumChildren(); i++) {
                  ((SalsaNode)children[i]).first_token.image = ", " +
                     ((SalsaNode)children[i]).first_token;
               }
            return "";
         case SalsaParserTreeConstants.JJTCONDITIONALEXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inToken("?", SalsaParserTreeConstants.JJTEXPRESSION);
               inToken(":", SalsaParserTreeConstants.JJTCONDITIONALEXPRESSION);
            }
            return ("");
         case SalsaParserTreeConstants.JJTCONDITIONALOREXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inPostToken("||", SalsaParserTreeConstants.JJTCONDITIONALANDEXPRESSION);
            }
            return ("");
         case SalsaParserTreeConstants.JJTCONDITIONALANDEXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inPostToken("&&", SalsaParserTreeConstants.JJTINCLUSIVEOREXPRESSION);
            }
            return ("");
         case SalsaParserTreeConstants.JJTINCLUSIVEOREXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inPostToken("|", SalsaParserTreeConstants.JJTEXCLUSIVEOREXPRESSION);
            }
            return ("");
         case SalsaParserTreeConstants.JJTEXCLUSIVEOREXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inPostToken("^", SalsaParserTreeConstants.JJTANDEXPRESSION);
            }
            return ("");
         case SalsaParserTreeConstants.JJTANDEXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inPostToken("&", SalsaParserTreeConstants.JJTEQUALITYEXPRESSION);
            }
            return ("");
         case SalsaParserTreeConstants.JJTEQUALITYEXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inPostToken(((SalsaNode)jjtGetChild(0)).last_token.next.image, // may work for all!
                     SalsaParserTreeConstants.JJTINSTANCEOFEXPRESSION);
            }
            return ("");
         case SalsaParserTreeConstants.JJTINSTANCEOFEXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inToken(" instanceof ",
                     SalsaParserTreeConstants.JJTTYPE);
            }
            return ("");
         case SalsaParserTreeConstants.JJTRELATIONALEXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inPostToken(((SalsaNode)jjtGetChild(0)).last_token.next.image,
                     SalsaParserTreeConstants.JJTSHIFTEXPRESSION);
            }
            return ("");
         case SalsaParserTreeConstants.JJTSHIFTEXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inPostToken(((SalsaNode)jjtGetChild(0)).last_token.next.image,
                     SalsaParserTreeConstants.JJTADDITIVEEXPRESSION);
            }
            return ("");
         case SalsaParserTreeConstants.JJTADDITIVEEXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inPostToken(((SalsaNode)jjtGetChild(0)).last_token.next.image,
                     SalsaParserTreeConstants.JJTMULTIPLICATIVEEXPRESSION);
            }
            return ("");
         case SalsaParserTreeConstants.JJTMULTIPLICATIVEEXPRESSION:
            if (jjtGetNumChildren() > 1) {
               inPostToken(((SalsaNode)jjtGetChild(0)).last_token.next.image,
                     SalsaParserTreeConstants.JJTUNARYEXPRESSION);
            }
            return ("");
         case SalsaParserTreeConstants.JJTUNARYEXPRESSION:
            if (((SalsaNode)jjtGetChild(0)).id == SalsaParserTreeConstants.JJTUNARYEXPRESSION)
               return (first_token.image);
            else
               return ("");
         case SalsaParserTreeConstants.JJTPREINCREMENTEXPRESSION:
            return ("++");
         case SalsaParserTreeConstants.JJTPREDECREMENTEXPRESSION:
            return ("--");
         case SalsaParserTreeConstants.JJTUNARYEXPRESSIONNOTPLUSMINUS:
            if (((SalsaNode)jjtGetChild(0)).id == SalsaParserTreeConstants.JJTUNARYEXPRESSION)
               return (first_token.image);
            else
               return ("");
         case SalsaParserTreeConstants.JJTCASTEXPRESSION:
            inToken(")", SalsaParserTreeConstants.JJTUNARYEXPRESSION);
            inToken(")", SalsaParserTreeConstants.JJTUNARYEXPRESSIONNOTPLUSMINUS);
            return ("(");
         case SalsaParserTreeConstants.JJTPRIMARYEXPRESSION:
            {
               MessageSendingData messageSendingData = getMessageSendingData();
               if (messageSend)
                  return SalsaCodeGenerator.generateMessageSend
                        (messageSendingData.getTarget(),
                        messageSendingData.getTargets(),
                        messageSendingData.getMethodNames(),
                        messageSendingData.getArguments(),
                        messageSendingData.getTokens(),
                        messageSendingData.getSender());
               else
                  return "";
            }
         case SalsaParserTreeConstants.JJTPRIMARYPREFIX:
            switch (first_token.kind) {
               case SalsaParserConstants.THIS:
                  return ("this");
               case SalsaParserConstants.SUPER:
                  return image();
               case SalsaParserConstants.LPAREN:
                  return "(";
               default:
                  return "";
            }
         case SalsaParserTreeConstants.JJTPRIMARYSUFFIX:
            switch (first_token.kind) {
               case SalsaParserConstants.DOT:
                  if (jjtGetNumChildren() == 0)
                     return image();
                  else
                     return ".";
               case SalsaParserConstants.LBRACKET:
                  return "[";
            }
         case SalsaParserTreeConstants.JJTLITERAL:
            if (jjtGetNumChildren() == 0)
               return image();
            else
               return "";
         case SalsaParserTreeConstants.JJTARGUMENTS:
            return "(";
         case SalsaParserTreeConstants.JJTARGUMENTLIST:
            return ("");
         case SalsaParserTreeConstants.JJTALLOCATIONEXPRESSION:
            return "new ";
         case SalsaParserTreeConstants.JJTARRAYDIMSANDINITS:
            if (((SalsaNode)jjtGetChild(0)).id == SalsaParserTreeConstants.JJTEXPRESSION) {
               return "[";
            } else { // ArrayInitializer
               Token token = first_token;
               StringBuffer str = new StringBuffer();
               while (token.image == "[") {
                  str.append("[]");
                  token = token.next.next;
               }
               return str.toString();
            }
         case SalsaParserTreeConstants.JJTBLOCKSTATEMENT:
            // cvarela: need to keep track of indentation,... like "dump".
            return "\t\t\t";
         case SalsaParserTreeConstants.JJTLOCALVARIABLEDECLARATION:
            // Modified by Travis Desell,  March 27, 2002

            if (jjtGetNumChildren() > 2)
               for (int i = 2; i < jjtGetNumChildren(); i++) {
                  ((SalsaNode)children[i]).first_token.image = ", " +
                     ((SalsaNode)children[i]).first_token;
               }
            return "";
         case SalsaParserTreeConstants.JJTEMPTYSTATEMENT:
            return "\t\t;";
         case SalsaParserTreeConstants.JJTSTATEMENTEXPRESSION:
            return "";
         case SalsaParserTreeConstants.JJTCONTINUATIONEXPRESSION:
            return generateContinuationExpressionCode(1, true, "", null);
         case SalsaParserTreeConstants.JJTLABELEDSTATEMENT:
            return "\n"+first_token.image+":";
         case SalsaParserTreeConstants.JJTSWITCHSTATEMENT:
            inToken("){", SalsaParserTreeConstants.JJTSWITCHLABEL);
            return "\nswitch (";
         case SalsaParserTreeConstants.JJTSWITCHLABEL:
            if (jjtGetNumChildren() == 0)
               return "\ndefault ";
            else // case
               return "\ncase ";
         case SalsaParserTreeConstants.JJTIFSTATEMENT:
            inToken(")\n\t\t\t", SalsaParserTreeConstants.JJTSTATEMENT);
            return "\n\t\tif (";
         case SalsaParserTreeConstants.JJTWHILESTATEMENT:
            inToken(")", SalsaParserTreeConstants.JJTSTATEMENT);
            return "\nwhile (";
         case SalsaParserTreeConstants.JJTDOSTATEMENT:
            inToken("while (", SalsaParserTreeConstants.JJTEXPRESSION);
            return "\ndo ";
         case SalsaParserTreeConstants.JJTFORSTATEMENT:
            //Modified by Travis Desell,  March 27, 2002

            switch ( ((SalsaNode)children[0]).id ) {
               case SalsaParserTreeConstants.JJTSTATEMENT:
                  return "\n\t\t\tfor (;;)";

               case SalsaParserTreeConstants.JJTFORUPDATE:
                  inToken(")", SalsaParserTreeConstants.JJTSTATEMENT);
                  return "\n\t\t\tfor (;;";

               case SalsaParserTreeConstants.JJTEXPRESSION:
                  switch ( ((SalsaNode)children[1]).id ) {
                     case SalsaParserTreeConstants.JJTSTATEMENT:
                        inToken(";)", SalsaParserTreeConstants.JJTSTATEMENT);
                        return "\n\t\t\tfor (;";

                     case SalsaParserTreeConstants.JJTFORUPDATE:
                        inToken(";", SalsaParserTreeConstants.JJTFORUPDATE);
                        inToken(")", SalsaParserTreeConstants.JJTSTATEMENT);
                        return "\n\t\t\tfor (;";
                  }

               case SalsaParserTreeConstants.JJTFORINIT:
                  switch ( ((SalsaNode)children[1]).id ) {
                     case SalsaParserTreeConstants.JJTEXPRESSION:
                        inToken(";", SalsaParserTreeConstants.JJTEXPRESSION);
                        switch ( ((SalsaNode)children[2]).id ) {
                           case SalsaParserTreeConstants.JJTFORUPDATE:
                              inToken(";", SalsaParserTreeConstants.JJTFORUPDATE);
                              inToken(")", SalsaParserTreeConstants.JJTSTATEMENT);
                              return "\n\t\t\tfor (";

                           case SalsaParserTreeConstants.JJTSTATEMENT:
                              inToken(";)", SalsaParserTreeConstants.JJTSTATEMENT);
                              return "\n\t\t\tfor (";
                        }

                     case SalsaParserTreeConstants.JJTFORUPDATE:
                        inToken(";;", SalsaParserTreeConstants.JJTFORUPDATE);
                        inToken(")", SalsaParserTreeConstants.JJTSTATEMENT);
                        return "\n\t\t\tfor (";

                     case SalsaParserTreeConstants.JJTSTATEMENT:
                        inToken(";;)", SalsaParserTreeConstants.JJTSTATEMENT);
                        return "\n\t\t\tfor (";
                  }
            }
         case SalsaParserTreeConstants.JJTSTATEMENTEXPRESSIONLIST:
            if (jjtGetNumChildren() > 1)
               inPostToken(",",SalsaParserTreeConstants.JJTSTATEMENTEXPRESSION);
            return ("");
         case SalsaParserTreeConstants.JJTBREAKSTATEMENT:
            if (jjtGetNumChildren() > 0)
               return "\nbreak "+((SalsaNode)jjtGetChild(0)).first_token.image;
            else
               return "\nbreak ";
         case SalsaParserTreeConstants.JJTCONTINUESTATEMENT:
            if (jjtGetNumChildren() > 0)
               return "\ncontinue "+((SalsaNode)jjtGetChild(0)).first_token.image;
            else
               return "\ncontinue ";
         case SalsaParserTreeConstants.JJTRETURNSTATEMENT:
            return "\n\t\t\treturn ";
         case SalsaParserTreeConstants.JJTTHROWSTATEMENT:
            return "\nthrow ";
         case SalsaParserTreeConstants.JJTSYNCHRONIZEDSTATEMENT:
            inToken(")", SalsaParserTreeConstants.JJTBLOCK);
            return ("\n synchronized (");
         case SalsaParserTreeConstants.JJTTRYSTATEMENT:
            // cvarela: support for "finally" must be included
            inToken("\t\tcatch (", SalsaParserTreeConstants.JJTFORMALPARAMETER);
            inPostToken(")", SalsaParserTreeConstants.JJTFORMALPARAMETER);
            return ("try");
         case SalsaParserTreeConstants.JJTJOINSTATEMENT:
            return SalsaCodeGenerator.generateJoinStatement
                  (getJoinStatementData(1,
                        true,
                        true));
         default:
            return "";
      }
   }

   public String postJavaCode() throws CodeGenerationException {
      StringBuffer tokenString = new StringBuffer();
      switch (id) {
         case SalsaParserTreeConstants.JJTMODULEDECLARATION:
            return (";\n\nimport salsa.language.*;\n");
         case SalsaParserTreeConstants.JJTIMPORTDECLARATION:
            // (".""*")?
            Token dot = first_token; // "import"
            Token semicolon = first_token.next.next;
            while (semicolon != last_token) { // ";"
               dot = dot.next;
               semicolon = semicolon.next;
            }
            if ((dot.kind == SalsaParserConstants.DOT)&&
                  (dot.next.kind == SalsaParserConstants.STAR))
               return ".*;\n";
            // no break here!!
         case SalsaParserTreeConstants.JJTFIELDDECLARATION:
         case SalsaParserTreeConstants.JJTBREAKSTATEMENT:
         case SalsaParserTreeConstants.JJTCONTINUESTATEMENT:
         case SalsaParserTreeConstants.JJTRETURNSTATEMENT:
         case SalsaParserTreeConstants.JJTTHROWSTATEMENT:
            return (";\n");
         case SalsaParserTreeConstants.JJTBLOCK:
            return ("\t\t}\n");
         case SalsaParserTreeConstants.JJTBEHAVIORBODY:


            String constructors = new String();

            // Determine which constructors were supplied.

            boolean defaultSupplied = false;

            for (int i=0; i < jjtGetNumChildren(); ++i) {

               // Check each behaviorbodydeclaration.
               SalsaNode node = (SalsaNode)jjtGetChild( i ).jjtGetChild( 0 );

               if (node.getId() == SalsaParserTreeConstants.JJTCONSTRUCTORDECLARATION) {

                  if (node.jjtGetChild( 0 ).jjtGetNumChildren() == 0) {
                     defaultSupplied = true;
                  }

                  String name = node.getFirstToken().image;

                  // Get all the tokens in the declaration.
                  String declarationArgs = "";

                  Token first = node.getFirstToken().next;
                  Token last = ((SalsaNode)node.jjtGetChild( 0 )).getLastToken().next;

                  do {
                     declarationArgs += first.image + " ";

                     first = first.next;

                  } while (first != last);

                  // Get all of the variabledeclaratorid's for State constructor invocation.
                  String invocationArgs = "";

                  // Iterate through all of the children of the FORMALPARAMETER child.
                  SalsaNode formalParameters = (SalsaNode)node.jjtGetChild( 0 );

                  for (int j=0; j<formalParameters.jjtGetNumChildren(); ++j) {

                     // Check each FormalParameter looking for VariableDeclaratorId's
                     SalsaNode declarator = (SalsaNode)(formalParameters.jjtGetChild( j ).jjtGetChild( 1 ));

                        first = declarator.getFirstToken();
                        last = declarator.getLastToken();

                        if (j > 0) {
                           invocationArgs += ", ";
                        }

                        do {
                           invocationArgs += first.image;
                        }
                        while (first != last);



                  }


                  // Copy the constructor body.
                  constructors += "\tpublic " + name + " " + declarationArgs + " {\n\n" +
                                    "\t\tselfReference = this;\n" +
                                    "\t\tactorState = new State( " + invocationArgs + " );\n\t}\n\n";
               }
            }

            String code = new String();

            if (!defaultSupplied) {
		// Default constructor for the state.
		code =
		"\t\tState() {\n" +
		"\t\t\tsuper.setSelf( selfReference );\n" +
		"\t\t\tthis.self = (" + SalsaCodeGenerator.behaviorName + ")selfReference;\n" +
		"\t\t}\n\n" +
		"\t\tState( UniversalActor _self ) {\n" +
		"\t\t\tsuper.setSelf( _self );\n" +
		"\t\t\tthis.self = (" + SalsaCodeGenerator.behaviorName + ")selfReference;\n" +	
		"\t\t}\n" +
		"\t}\n\n" +
		"\tpublic "+SalsaCodeGenerator.behaviorName+"() { \n" +
		"\t\tselfReference = this;\n" +
		"\t\tactorState = new State();\n" +
		"\t}\n\n" +
		constructors + "\n\n";
	    }

            else {
               code =
               "\t}\n\n" + constructors + "\n\n";
            }

            return code + SalsaCodeGenerator.generateMainMethod() + "\n}\n";

         case SalsaParserTreeConstants.JJTSWITCHSTATEMENT:
            return ("}\n");
         case SalsaParserTreeConstants.JJTARRAYINITIALIZER:
            return ("}");
         case SalsaParserTreeConstants.JJTMETHODDECLARATION:
            if (last_token.kind == SalsaParserConstants.SEMICOLON)
               return (";");
            else
               return "\n";
         case SalsaParserTreeConstants.JJTFORMALPARAMETERS:
            return (")");
         case SalsaParserTreeConstants.JJTFORMALPARAMETER:
            ifLastOutPostTokenNR(SalsaParserTreeConstants.JJTFORMALPARAMETER);
            return ("");
         case SalsaParserTreeConstants.JJTCONSTRUCTORDECLARATION:
            outAllTokens();

            if (jjtGetNumChildren() == 1) // Only FormalParameters
               return "{\n\t\t\tsuper.setSelf( selfReference );\n"+
                      "\t\t\tthis.self = (" + SalsaCodeGenerator.behaviorName + ")selfReference;\n\t\t}\n\n";
               
            return "\t\t\tsuper.setSelf( selfReference );\n"+
                   "\t\t\tthis.self = (" + SalsaCodeGenerator.behaviorName + ")selfReference;\n\t\t}\n\n";

         case SalsaParserTreeConstants.JJTEXPLICITCONSTRUCTORINVOCATION:
            ((SalsaNode)jjtGetParent()).
                  outToken(SalsaParserTreeConstants.JJTBLOCKSTATEMENT);
            outToken(SalsaParserTreeConstants.JJTARGUMENTS);
            return (";");
         case SalsaParserTreeConstants.JJTTYPE:
         if (first_token != last_token) {  // ("[""]")*
            String typeString = "";
            Token tok = first_token;
            while ((tok = tok.next) != last_token)
               if ((tok.kind == SalsaParserConstants.LBRACKET)||
                     (tok.kind == SalsaParserConstants.RBRACKET))
               typeString += tok.image;
            if (tok.kind == SalsaParserConstants.RBRACKET)
               typeString += tok.image;
            return typeString+" ";
         } else
            return " ";
         case SalsaParserTreeConstants.JJTRESULTTYPE:
            return (" ");
         case SalsaParserTreeConstants.JJTNAMELIST:
            outToken(SalsaParserTreeConstants.JJTNAME);
            return ("");
         case SalsaParserTreeConstants.JJTEXPRESSION:
            if (((SalsaNode)jjtGetParent()).id ==
                  SalsaParserTreeConstants.JJTARGUMENTLIST)
               ((SalsaNode)jjtGetParent()).
                     inToken(", ",SalsaParserTreeConstants.JJTEXPRESSION);
            return "";
         case SalsaParserTreeConstants.JJTCONDITIONALANDEXPRESSION:
            ifLastOutPostToken(SalsaParserTreeConstants.JJTCONDITIONALANDEXPRESSION);
            return ("");
         case SalsaParserTreeConstants.JJTINCLUSIVEOREXPRESSION:
            ifLastOutPostToken(SalsaParserTreeConstants.JJTINCLUSIVEOREXPRESSION);
            return ("");
         case SalsaParserTreeConstants.JJTEXCLUSIVEOREXPRESSION:
            ifLastOutPostToken(SalsaParserTreeConstants.JJTEXCLUSIVEOREXPRESSION);
            return ("");
         case SalsaParserTreeConstants.JJTANDEXPRESSION:
            ifLastOutPostToken(SalsaParserTreeConstants.JJTANDEXPRESSION);
            return ("");
         case SalsaParserTreeConstants.JJTEQUALITYEXPRESSION:
            ifLastOutPostToken(SalsaParserTreeConstants.JJTEQUALITYEXPRESSION);
            return ("");
         case SalsaParserTreeConstants.JJTINSTANCEOFEXPRESSION:
            ifLastOutPostToken(SalsaParserTreeConstants.JJTINSTANCEOFEXPRESSION);
            return ("");
         case SalsaParserTreeConstants.JJTSHIFTEXPRESSION:
            ifLastOutPostToken(SalsaParserTreeConstants.JJTSHIFTEXPRESSION);
            return ("");
         case SalsaParserTreeConstants.JJTADDITIVEEXPRESSION:
            ifLastOutPostToken(SalsaParserTreeConstants.JJTADDITIVEEXPRESSION);
            return ("");
         case SalsaParserTreeConstants.JJTMULTIPLICATIVEEXPRESSION:
            ifLastOutPostToken(SalsaParserTreeConstants.JJTMULTIPLICATIVEEXPRESSION);
            return ("");
         case SalsaParserTreeConstants.JJTUNARYEXPRESSION:
            ifLastOutPostToken(SalsaParserTreeConstants.JJTUNARYEXPRESSION);
            return ("");
         case SalsaParserTreeConstants.JJTSTATEMENTEXPRESSION:
            ifLastOutPostToken(SalsaParserTreeConstants.JJTSTATEMENTEXPRESSION);
            // no break here!!!
         case SalsaParserTreeConstants.JJTPOSTFIXEXPRESSION:
            if (last_token.kind == SalsaParserConstants.INCR)
               return ("++");
            else if (last_token.kind == SalsaParserConstants.DECR)
               return ("--");
            else
               return "";
         case SalsaParserTreeConstants.JJTPRIMARYEXPRESSION:
            return "";
         case SalsaParserTreeConstants.JJTPRIMARYPREFIX:
            if (first_token.kind==SalsaParserConstants.LPAREN)
               return ")";
            else if ((jjtGetNumChildren()>0) &&
                  ((SalsaNode)jjtGetChild(0)).id ==
                  SalsaParserTreeConstants.JJTRESULTTYPE)
               return ".class";
            else
               return "";
         case SalsaParserTreeConstants.JJTPRIMARYSUFFIX:
            if (first_token.kind==SalsaParserConstants.LBRACKET)
               return "]";
            else
               return "";
         case SalsaParserTreeConstants.JJTARGUMENTS:
            return ")";
         case SalsaParserTreeConstants.JJTSTATEMENT:
            SalsaNode parent = (SalsaNode)jjtGetParent();
            if (parent.id == SalsaParserTreeConstants.JJTIFSTATEMENT) {
               parent.outToken(SalsaParserTreeConstants.JJTSTATEMENT);
               if (parent.jjtGetNumChildren() > 2) {
                  parent.inToken("\t\telse\n\t\t\t",
                        SalsaParserTreeConstants.JJTSTATEMENT);
               }
            }
            if (((SalsaNode)jjtGetChild(0)).id ==
                  SalsaParserTreeConstants.JJTSTATEMENTEXPRESSION)
               return ";\n";
            else
               return "";
         case SalsaParserTreeConstants.JJTARRAYDIMSANDINITS:
            if (((SalsaNode)jjtGetChild(0)).id ==
                  SalsaParserTreeConstants.JJTEXPRESSION) {
               StringBuffer str = new StringBuffer("]");
               Token token = ((SalsaNode)jjtGetChild(0)).last_token.next.next;
               while (token.image == "[") {
                  str.append("[]");
                  token = token.next.next;
               }
               return str.toString();
            } else  // ArrayInitializer
               return "";
         case SalsaParserTreeConstants.JJTBLOCKSTATEMENT:
            ((SalsaNode)jjtGetParent()).
                  outToken(SalsaParserTreeConstants.JJTBLOCKSTATEMENT);
            if (((SalsaNode)jjtGetChild(0)).id ==
                  SalsaParserTreeConstants.JJTLOCALVARIABLEDECLARATION)
               return ";";
            else
               return "";
         case SalsaParserTreeConstants.JJTSWITCHLABEL:
            ((SalsaNode)jjtGetParent()).
                  outToken(SalsaParserTreeConstants.JJTSWITCHLABEL);
            return ":\n";
         case SalsaParserTreeConstants.JJTDOSTATEMENT:
            return ");";
         default:
            return "";
      }
   }

   public void generateJavaCode(java.io.PrintWriter writer)
         throws CodeGenerationException {
      if (generateCode) {
         String preJavaCode = preToken(id)+preJavaCode();
         writer.print(preJavaCode);
         if (children != null) {
            for (int i = 0; i < children.length; ++i) {
               SalsaNode n = (SalsaNode)children[i];
               if (n != null) {
                  n.generateJavaCode(writer);
               }
            }
         }
         String postJavaCode = postToken(id)+postJavaCode();
         writer.print(postJavaCode);
      }
   }

   public String toJavaCode()
         throws CodeGenerationException {
      StringBuffer javaCode = new StringBuffer();
      String preJavaCode = preToken(id)+preJavaCode();
      javaCode.append(preJavaCode);
      if (children != null) {
         for (int i = 0; i < children.length; ++i) {
            SalsaNode n = (SalsaNode)children[i];
            if (n != null) {
               javaCode.append(n.toJavaCode());
            }
         }
      }
      String postJavaCode = postToken(id)+postJavaCode();
      javaCode.append(postJavaCode);
      return javaCode.toString();
   }

   public String preToken(int id) {
      String preTokenString;
      SalsaNode parentNode = (SalsaNode) jjtGetParent();
      if ((parentNode != null) &&
            (preTokenString = (String) parentNode.preTokens.get(new Integer(id)))
            != null)
         return preTokenString;
      else
         return "";
   }

   public void inToken(String pre, int id) {
      preTokens.put(new Integer(id), pre);
   }

   public void outToken(int id) {
      preTokens.remove(new Integer(id));
   }

   public void outAllTokens() {
      preTokens.clear();
   }

   public String postToken(int id) {
      String postTokenString;
      SalsaNode parentNode = (SalsaNode) jjtGetParent();
      if ((parentNode != null) &&
            (postTokenString = (String) parentNode.postTokens.get(new Integer(id)))
            != null)
         return postTokenString;
      else
         return "";
   }

   public void inPostToken(String pre, int id) {
      postTokens.put(new Integer(id), pre);
   }

   public void outPostToken(int id) {
      postTokens.remove(new Integer(id));
   }

   public void outAllPostTokens() {
      postTokens.clear();
   }

   public void ifLastOutPostToken(int id) {
      // if previous to last child, no more post tokens
      SalsaNode parent = (SalsaNode) jjtGetParent();
      int numChildren = parent.jjtGetNumChildren();
      if (numChildren > 1)
         if (this == parent.jjtGetChild(numChildren-2))
            parent.outPostToken(id);
         else  // replace to operator between subsequent nodes
            for (int i = 0; i < numChildren-2; i++) {
               if (this == parent.jjtGetChild(i)) {
                  ((SalsaNode)jjtGetParent()).
                        inPostToken(((SalsaNode)parent.jjtGetChild(i+1)).
                        last_token.next.image, id);
                  break;
               }
            }
   }

   public void ifLastOutPostTokenNR(int id) {
      // same as above but No Replacement of tokens.
      SalsaNode parent = (SalsaNode) jjtGetParent();
      int numChildren = parent.jjtGetNumChildren();
      if (numChildren > 1)
         if (this == parent.jjtGetChild(numChildren-2))
            parent.outPostToken(id);
   }

   public MessageSendingData getMessageSendingData()
         throws CodeGenerationException {
      // only for PrimaryExpressions
      if (id != SalsaParserTreeConstants.JJTPRIMARYEXPRESSION)
         throw new CodeGenerationException("Salsa Error:  getMessageSendingData"
               +" must receive a PRIMARYEXPRESSION node\n");
      // Select OUT Salsa Actor message sending syntax:
      // (1)  No PrimarySuffix, or
      // (2)  Only ONE PrimarySuffix of type Arguments, or
      // (3)  A MessageSendSuffix as last non-terminal
      String target = null,
      targets = null,
      methodNames = null,
      arguments = null,
      tokens = null,
      sender = null;
      messageSend = false;
      int numChildren = jjtGetNumChildren();
      if ((numChildren == 1) &&
            (((SalsaNode)jjtGetParent()).id ==
            SalsaParserTreeConstants.JJTCONTINUATIONEXPRESSION)) {
         // m ::= m(token)
         target = targets = sender = "self";
         methodNames = "\""+image()+"\"";
         arguments = "{ new Token() }";
         tokens = "0";
         messageSend = true;
         ((SalsaNode)jjtGetChild(0)).generateCode = false;
      } else if ((numChildren == 2) &&
            (((SalsaNode)jjtGetChild(1)).id ==
            SalsaParserTreeConstants.JJTPRIMARYSUFFIX) &&
            (jjtGetChild(1).jjtGetNumChildren() > 0) &&
            (((SalsaNode)(jjtGetChild(1).jjtGetChild(0))).id ==
            SalsaParserTreeConstants.JJTARGUMENTS)) {
         // o.m() is selected OUT:
         if (((SalsaNode)(jjtGetChild(0))).image().indexOf(".") == -1) {
            // m()  or  m(arg1,arg2,..)
            target = targets = sender = "self";
            methodNames = "\""+((SalsaNode)jjtGetChild(0)).image()+"\"";
            MessageSendingData argumentsAndTokens =
                  ((SalsaNode)(jjtGetChild(1).jjtGetChild(0))).
                  getArgumentsAndTokens();
            arguments = argumentsAndTokens.getArguments();
            tokens = argumentsAndTokens.getTokens();
            messageSend = true;
            ((SalsaNode)jjtGetChild(0)).generateCode = false;
            ((SalsaNode)jjtGetChild(1)).generateCode = false;
         }
      } else if (((SalsaNode)jjtGetChild(numChildren-1)).id ==
            SalsaParserTreeConstants.JJTMESSAGESENDSUFFIX){
         // a:m   or   a:m(args)
         target = ((SalsaNode)jjtGetChild(0)).image();

         if ("this".equals( target )) {
            target = "self";
         }

         ((SalsaNode)jjtGetChild(0)).generateCode = false;
         for (int i = 1; i < numChildren - 1; i++) {
            target = target + ((SalsaNode)jjtGetChild(i)).image();
            ((SalsaNode)jjtGetChild(i)).generateCode = false;
         }
         int i = numChildren - 1;
         sender = "self";
         targets = target;
         methodNames = "\""+
         ((SalsaNode)jjtGetChild(i)).first_token.next.image + "\"";
         ((SalsaNode)jjtGetChild(i)).generateCode = false;
         if ( jjtGetChild(i).jjtGetNumChildren() == 0 ) {
            // a:m  ::= a:m(token)
            arguments = "{ new Token() }";
            tokens = "0";
         } else {
            // a:m()   or   a:m(args)
            MessageSendingData argumentsAndTokens =
                  ((SalsaNode)(jjtGetChild(i).jjtGetChild(0))).
                  getArgumentsAndTokens();
            arguments = argumentsAndTokens.getArguments();
            tokens = argumentsAndTokens.getTokens();
            ((SalsaNode)(jjtGetChild(i).jjtGetChild(0))).generateCode = false;
         }
         messageSend = true;
      }
      if (messageSend)
         return new MessageSendingData
               (sender, target, targets, methodNames, arguments, tokens);
      else
         return null;
      }
      MessageSendingData getArgumentsAndTokens()
            throws CodeGenerationException{
      // only for Arguments nodes
      if (id != SalsaParserTreeConstants.JJTARGUMENTS)
         throw new CodeGenerationException("Salsa Error:  getArgumentsAndTokens"
               +" must receive an ARGUMENTS node\n");

      // When literals are found without context as arguments,
      // they get wrapped to Objects.  If a token is found without
      // context, it gets changed to an instance of the Token class
      // and its argument position
      // is recorded.
      if (jjtGetNumChildren() == 0) // ()
         return new MessageSendingData("{}", "-1");
      else {
         String arguments = "{";
         String tokens = "-1";
         SalsaNode argumentList = (SalsaNode)jjtGetChild(0);
         SalsaNode singleLiteral = null;
         if (argumentList.children != null) {
            for (int i = 0; i < argumentList.children.length; ++i) {
               SalsaNode argument = (SalsaNode)argumentList.children[i];
               if (argument != null) {
                  // for each Argument
                  if ((singleLiteral = argument.getSingleLiteral()) != null) {
                     // argument is a single Literal
                     if (i!=0)
                        arguments +=", ";  // preJavaCode()
                     switch (singleLiteral.id) {
                        case SalsaParserTreeConstants.JJTINTEGERLITERAL:
                           // cvarela: should check for Long/Byte/Short
                           arguments += "new Integer("+singleLiteral.image()+")";
                           break;
                        case SalsaParserTreeConstants.JJTFLOATINGPOINTLITERAL:
                           // cvarela: should check for Double
                           arguments += "new Float("+singleLiteral.image()+")";
                           break;
                        case SalsaParserTreeConstants.JJTCHARACTERLITERAL:
                           arguments += "new Character("+singleLiteral.image()+")";
                           break;
                        case SalsaParserTreeConstants.JJTBOOLEANLITERAL:
                           arguments += "new Boolean("+singleLiteral.image()+")";
                           break;
                        case SalsaParserTreeConstants.JJTTOKENLITERAL:
                           arguments += "new Token()";
                           tokens = i+"";
                           break;
                        default:  // null and String literals remain unchanged.
                           arguments += singleLiteral.image();
                     }
                     // postJavaCode()
                     argumentList.inToken(", ",SalsaParserTreeConstants.JJTEXPRESSION);
                  } else {
                     arguments += argument.toJavaCode();
                  }
               }
            }
         }
         arguments+="}";
         return new MessageSendingData(arguments, tokens);
      }
   }

   SalsaNode getSingleLiteral() {
      // only for Expression nodes, or their children
      // if the expression is a single literal, return such literal,
      // else returns null
      int numChildren = jjtGetNumChildren();
      if (numChildren == 0)
         // We have a single literal now!
         return this;
      else if (numChildren > 1)
         // If more than one child, this expression is not a single literal
         return null;
      else
         // We have exactly one child, keep trying!
         return ((SalsaNode)jjtGetChild(0)).getSingleLiteral();
   }

   String generateContinuationExpressionCode(int number, boolean withSend,
         String messageName, String tokenClass)
         throws CodeGenerationException {
      // only for ContinuationExpression
      if (id != SalsaParserTreeConstants.JJTCONTINUATIONEXPRESSION)
         throw new CodeGenerationException("Salsa Error:  "
               +"generateContinuationExpressionCode"
               +" must receive a CONTINUATIONEXPRESSION"
               +" node\n");
      if (jjtGetNumChildren() == 0) {  // "currentContinuation"
         return "";
      } else if (((SalsaNode)jjtGetChild(0)).id ==
            SalsaParserTreeConstants.JJTPRIMARYEXPRESSION) {
         MessageSendingData messageSendingData =  new MessageSendingData();
         SalsaNode continuationExpression = this;
         while (continuationExpression.jjtGetNumChildren() == 2) {
            // PE -> CE
            SalsaNode primaryExpression =
                  (SalsaNode)continuationExpression.jjtGetChild(0);
            primaryExpression.generateCode = false;
            messageSendingData.append
                  (primaryExpression.getMessageSendingData());
            continuationExpression =
                  (SalsaNode)continuationExpression.jjtGetChild(1);
            continuationExpression.generateCode = false;
         }  // PE || "currentContinuation" || JoinStatement
         if (continuationExpression.jjtGetNumChildren() == 0) {
         // "currentContinuation"
            return "";
         } else { // numchildren == 1
            SalsaNode node =
               (SalsaNode)continuationExpression.jjtGetChild(0);
            if (node.id == SalsaParserTreeConstants.JJTPRIMARYEXPRESSION) {
               // PE
               SalsaNode primaryExpression = node;
               primaryExpression.generateCode = false;
               messageSendingData.append
                     (primaryExpression.getMessageSendingData());
               if (!withSend) messageSendingData.setTarget(messageName);
                  if (tokenClass != null)
                     messageSendingData.setTokenClass(tokenClass);
               return SalsaCodeGenerator.generateMessageSend
                     (messageSendingData, withSend);
            } else { // JoinStatement
               // (1) generate JS definition code
               String joinStatementCode =
                     SalsaCodeGenerator.generateJoinStatement
                     (node.getJoinStatementData(number, false, false));
               // (2) link "director<-process()" to messageSendingData
               messageSendingData.append
                     (new MessageSendingData("this", "_joinDirector"+number,
                     "_joinDirector"+number, "\"process\"",
                     "{ }", "-1"));
               if (!withSend)
                  messageSendingData.setTarget(messageName);
               if (tokenClass != null)
                  messageSendingData.setTokenClass(tokenClass);
               return (joinStatementCode
                     +SalsaCodeGenerator.generateMessageSend
                     (messageSendingData, withSend));
            }
         }
      } else { // JoinStatement
         ((SalsaNode)jjtGetChild(0)).generateCode = false;
         return
               SalsaCodeGenerator.generateJoinStatement
               (((SalsaNode)jjtGetChild(0)).
               getJoinStatementData(number, withSend, true));
      }
   }

   public JoinStatementData getJoinStatementData(int number, boolean withSend,
         boolean withContinuation)
         throws CodeGenerationException {
      // only for JoinStatement
      if (id != SalsaParserTreeConstants.JJTJOINSTATEMENT)
         throw new CodeGenerationException("Salsa Error:  getJoinStatementData"
               +" must receive a JOINSTATEMENT node\n");

      // Select OUT Salsa Join Statement syntax:
      // (1)  from ContinuationExpressionList
      // (2)  from ContinuationExpression
      String sender = null,
      target = null,
      targets = null,
      methodNames = null,
      arguments = null,
      tokens = null;
      boolean actorArray;
      MessageSendingData msData;
      SalsaNode continuationExpressionList = (SalsaNode) jjtGetChild(0);
      SalsaNode continuationExpression = (SalsaNode) jjtGetChild(1);
      continuationExpressionList.generateCode = false;
      continuationExpression.generateCode = false;
      sender = "this";
      target = "_joinDirector"+number;
      int numChildren = continuationExpressionList.jjtGetNumChildren();
      if (numChildren == 1) {
         // array of actors
         actorArray = true;
         msData = ((SalsaNode) continuationExpressionList.jjtGetChild(0).  // CE
               jjtGetChild(0)).  // PE
               getMessageSendingData();
      } else {
         // actors are spelled out
         actorArray = false;
         msData = new MessageSendingData();
         for (int i=0; i<numChildren; i++) {
            SalsaNode joinExpression =
                  (SalsaNode) continuationExpressionList.jjtGetChild(i);
            msData.append(joinExpression.getJoinDirectorData());
         }
      }
      return new JoinStatementData
            (sender,
            target,
            msData.getTargets(),
            msData.getMethodNames(),
            msData.getArguments(),
            msData.getTokens(),
            continuationExpression,
            number,
            withSend,
            actorArray,
            withContinuation);
   }

   MessageSendingData getJoinDirectorData()
         throws CodeGenerationException {
      // only for ContinuationExpression
      if (id != SalsaParserTreeConstants.JJTCONTINUATIONEXPRESSION)
         throw new CodeGenerationException("Salsa Error:  "
               +"SalsaNode.getJoinDirectorData"
               +" must receive a CONTINUATIONEXPRESSION"
               +" node\n");
      if (jjtGetNumChildren() == 0)  // "currentContinuation"
         throw new InvalidJoinException("Salsa Error:  "
               +"currentContinuation not allowed "
               +"inside Join Statement\n");
      if (((SalsaNode)jjtGetChild(0)).id ==
            SalsaParserTreeConstants.JJTPRIMARYEXPRESSION)
         if (jjtGetNumChildren() == 2)
         // PE -> CE,   not dealt with yet.
            throw new InvalidJoinException("Salsa Error:  "
                  +"No @ continuations allowed inside "
                  +"Join Statement\n");
      else { // PE
         SalsaNode primaryExpression = (SalsaNode)jjtGetChild(0);
         primaryExpression.generateCode = false;
         return (primaryExpression.getMessageSendingData());
      }
      // JoinStatement
      throw new InvalidJoinException("Salsa Error:  "
            +"No joins allowed "
            +"inside Join Statement\n");
   }
}
